#include <stdio.h>
#include <assert.h>
#define epsilon (float)1e-5
#define DATA float
#define THREADxBLOCKalongXorY 32

void MatrixMulOnHost(DATA* M, DATA* N, DATA* P, int Width) {
    for (int i = 0; i < Width; ++i) {
        for (int j = 0; j < Width; ++j) {
            DATA pvalue = 0;
            for (int k = 0; k < Width; ++k) {
                DATA a = M[i * Width + k];
                DATA b = N[k * Width + j];
                pvalue += a * b;
            }
        P[i * Width + j] = pvalue;
        }
    }
}

__global__ void MatrixMulKernel(DATA* dM, DATA* dN, DATA* dP, int Width) {
   DATA Pvalue = 0.0F;
   int ix   = blockIdx.x * blockDim.x + threadIdx.x;
   int iy   = blockIdx.y * blockDim.y + threadIdx.y;
   
   if(ix < Width && iy < Width) {
#pragma unroll 4
      for (int k = 0; k < Width; ++k) {
       //  Pvalue += dM[iy*Width+k]*dN[k*Width + ix];
	 Pvalue += dM[ix*Width+k]*dN[k*Width + iy];
      }
      //dP[iy*Width+ix] = Pvalue;
      dP[ix*Width+iy] = Pvalue;
   }
}

__global__ void MatrixMulKernel_slow(DATA* dM, DATA* dN, DATA* dP, int Width) {

   int ix   = blockIdx.x * blockDim.x + threadIdx.x;
   int iy   = blockIdx.y * blockDim.y + threadIdx.y;

   int idx=iy*Width+ix;
   if(ix < Width && iy < Width) {
      dP[idx] = 0.0F;
      for (int k = 0; k < Width; ++k) {
         dP[idx] += dM[iy*Width+k]*dN[k*Width + ix];
      }
   }
}


void MatrixMulOnDevice(DATA* M, DATA* N, DATA* P, int Width) {

    int size = Width * Width * sizeof(DATA);
    float mflops;
    DATA *dM, *dN, *dP;

//  Gestione delle griglia ottimizzata per ogni dimensione di blocco
    int gridsize = Width/THREADxBLOCKalongXorY;
    if(gridsize*THREADxBLOCKalongXorY<Width) {
       gridsize=gridsize+1;
    }
    dim3 dimGrid(gridsize,gridsize);
    dim3 dimBlock(THREADxBLOCKalongXorY, THREADxBLOCKalongXorY);
    printf("Gridsize: %d\n", gridsize);

    cudaMalloc(&dM, size);
    cudaMemcpy(dM, M, size, cudaMemcpyHostToDevice);
    cudaMalloc(&dN, size);
    cudaMemcpy(dN, N, size, cudaMemcpyHostToDevice);
    cudaMalloc(&dP, size);

//  Calcolo mFlop totali, creazione eventi e record di event start
    mflops = 2.*Width*Width*Width ;
    cudaEvent_t start, stop;
    cudaEventCreate(&start);
    cudaEventCreate(&stop);
    cudaEventRecord(start, 0);

//  Chiamata cudaGetLastError per azzerare eventuali precedenti errori
    cudaError_t mycudaerror  ;
    mycudaerror = cudaGetLastError()  ;

//    cudaFuncSetCacheConfig(MatrixMulKernel,cudaFuncCachePreferL1);

 // Chiamata kernel
    MatrixMulKernel<<<dimGrid, dimBlock>>>(dM, dN, dP, Width);

//  Chiamata cudaGetLastError preceduta da sincronizzazione per rilevare eventali errori
    cudaThreadSynchronize() ; 
    mycudaerror = cudaGetLastError() ;
    if(mycudaerror != cudaSuccess)  {
      fprintf(stderr,"%s\n",cudaGetErrorString(mycudaerror)) ;
      exit(1);
    }

//  Record evento di stop, sincronizzazione eventi, calcolo elapsed time (e distruzione eventi)
    cudaEventRecord(stop, 0);
    cudaEventSynchronize(stop);
    float elapsed;
    cudaEventElapsedTime(&elapsed, start, stop);
    elapsed = elapsed/1000.f;
    cudaEventDestroy(start);
    cudaEventDestroy(stop);

//  Stampa tempi
    printf("Kernel elapsed time %fs \n", elapsed);
    mflops = mflops/(1000.f*1000.f*elapsed) ;
    printf("Mflops: %f", mflops);

// Copia P dalla memoria GPU
    cudaMemcpy(P, dP, size, cudaMemcpyDeviceToHost);

// Libera la memoria utilizzata per le matrici
    cudaFree(dM);
    cudaFree(dN);
    cudaFree(dP);

}

// main
int main(int argc, char** argv) {

  int Width;

  DATA *M, *N, *hP, *gP;

  if(argc<2) {
     fprintf(stderr,"Usage: %s Width\n",argv[0]);
     exit(1);
  }

  Width=atoi(argv[1]);

  if(Width<1) {
     fprintf(stderr,"Error Width=%d, must be > 0\n",Width);
     exit(1);

  }

  M=(DATA *)malloc(Width*Width*sizeof(DATA));
  N=(DATA *)malloc(Width*Width*sizeof(DATA));
  hP=(DATA *)malloc(Width*Width*sizeof(DATA));
  gP=(DATA *)malloc(Width*Width*sizeof(DATA));

  if(M==NULL) {
    fprintf(stderr,"Could not get memory for M\n");
    exit(1);
  }
  if(N==NULL) {
    fprintf(stderr,"Could not get memory for N\n");
    exit(1);
  }
  if(hP==NULL) {
    fprintf(stderr,"Could not get memory for hP\n");
    exit(1);
  }
  if(gP==NULL) {
    fprintf(stderr,"Could not get memory for gP\n");
    exit(1);
  }

  memset(gP,0,Width*Width*sizeof(DATA));
  memset(hP,0,Width*Width*sizeof(DATA));

  for(int y=0; y<Width; y++){
    for(int x=0; x<Width; x++) {
       M[y*Width+x]=(DATA)(((y+1)*Width+x+1)/(Width));
       N[y*Width+x]=(DATA)(((y+1)*Width+x+1)/(Width));
    }
  }

  MatrixMulOnHost(M, N, hP, Width);

  MatrixMulOnDevice(M, N, gP, Width);

  int errCnt = 0;
  for(int y=0; y<Width; y++){
   for(int x=0; x<Width; x++) {
      DATA it = hP[y*Width+x];
      if(fabs(it - gP[y*Width+x])> epsilon*it) {
       printf("failing x=%d, y=%d: %f!=%f \n",x,y,it,gP[y*Width+x]);
       errCnt++;
      }
   }
  }

  if(errCnt==0)
    printf("\nTEST PASSED\n");
  else
    printf("\n\nTEST FAILED: number of errors:  %d\n", errCnt);

}
