// includes
#include <stdio.h>

void checkCUDAError(const char *msg)
{
	cudaError_t err = cudaGetLastError();
	if( cudaSuccess != err)
	{
		printf( "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
		exit(-1);
	}
}

__global__ void fillArrayKernel(int *d_a)
{
	int idx  = blockIdx.x*blockDim.x + threadIdx.x;
	int idy  = blockIdx.y*blockDim.y + threadIdx.y;
	int ny=gridDim.y * blockDim.y;
	d_a[idx*ny+idy] = idx*1000+idy ;
}


__host__ void fillArrayHost(int *h_a,int nElemX,int nElemY){
	for (int i=0; i<nElemX; i++)
	{
		for (int j=0; j<nElemY; j++)
		{
			h_a[i*nElemY+j]=i*1000+j;
		}
	}
}


////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv)
{
	////////////////////////  Declarations ////////////////////////

	// pointers for host memory
	int *h_a_cpu; //pointer for CPU results
	int *h_a_gpu; //pointers for GPU results on host

	// pointer for device memory
	int *d_a;//pointers for GPU results on device

	// define grid and block sizes
	int numBlocksX = 16;
	int numBlocksY=  16;
	int numThreadsPerBlockX = 16;
	int numThreadsPerBlockY = 16;

	// number of elements and memory size
	int nElemX=numBlocksX*numThreadsPerBlockX;
	int nElemY=numBlocksY*numThreadsPerBlockY;
	
	
	size_t memSize = nElemX*nElemY * sizeof(int);
	

	////////////////////////  Allocate Memory   ////////////////////////

	//Allocate host memory
	h_a_cpu = (int *) malloc(memSize);
	h_a_gpu = (int *) malloc(memSize);


	//Allocate device memory
	cudaMalloc( (void **) &d_a, memSize );
	// Check for any CUDA errors
	checkCUDAError("cudaMalloc");

	////////////////////////  Calculation (host)  ////////////////////////
	fillArrayHost(h_a_cpu,nElemX,nElemY);

	///////////////////////  Calculation (Device)            /////////////////////////
	dim3 dimGrid(numBlocksX,numBlocksY);
	dim3 dimBlock(numThreadsPerBlockX,numThreadsPerBlockY);
	fillArrayKernel<<< dimGrid, dimBlock >>>( d_a );

	// block until the device has completed
	cudaThreadSynchronize();

	// check if kernel execution generated an error
	checkCUDAError("kernel execution");

	////////////////////////  Copy Data From Device To Host   ////////////////////////
	cudaMemcpy( h_a_gpu, d_a, memSize, cudaMemcpyDeviceToHost );
	// Check for any CUDA errors
	checkCUDAError("cudaMemcpyDeviceToHost");


	////////////////////////  Compare the results     ////////////////////////
	for (int i = 0; i < nElemX; i++)
	{
		for (int j = 0; j < nElemY; j++)
		{
			if(h_a_gpu[i*nElemY+j] != h_a_cpu[i*nElemY+j])
			{
				printf ("Error : i= %d j= %d h_a_gpu[i*nElemY+j]= %d h_a_cpu[i*nElemY+j]= %d \n", i, j, h_a_gpu[i*nElemY+j], h_a_cpu[i*nElemY+j]);
				printf("\nPress ENTER to exit...\n");
				getchar();
				return 1;
			}
		}
	}
	printf("No Error \n");

	////////////////////////  Free Memory      ////////////////////////
	// free device memory
	cudaFree(d_a);

	// free host memory
	free(h_a_cpu);
	free(h_a_gpu);

	////////////////////////  End    ////////////////////////
	//printf("Correct!\n");
	printf("\nPress ENTER to exit...\n");
	getchar();

	return 0;
}

