//////////////////////////////////////////////////////////////////////////////
// CUDA exercise 2: Mandelbrot set
//////////////////////////////////////////////////////////////////////////////
// includes
#include <stdio.h>
#include <cutil.h>
typedef  unsigned int  uint;
#define REAL double 
extern "C"{
#include "bmp_int.h"
}

void checkCUDAError(const char *msg)
{
	cudaError_t err = cudaGetLastError();
	if (cudaSuccess != err)
	{
		printf("Cuda error: %s: %s.\n", msg, cudaGetErrorString(err));
		printf("\nPress ENTER to exit...\n");
		getchar();
		exit(-1);
	}
}
__global__ void mandelbrotKernel(REAL xmin, REAL xmax, int nElemX, REAL ymin, REAL ymax, int nElemY, int niter, uint* d_a)
{
	int i = threadIdx.x + blockIdx.x * blockDim.x;
	int j = threadIdx.y + blockIdx.y * blockDim.y;
		REAL xc = xmin + (xmax - xmin) / (nElemX - 1) * i; //xc=real(c)
			REAL yc = ymin + (ymax - ymin) / (nElemY - 1) * j; //yc=imag(c)
			REAL x = 0.0; //x=real(z_k)
			REAL y = 0.0; //y=imag(z_k)
			for (int k = 0; k < niter; k = k + 1) //iteration loop
			{
				REAL tempx = x * x - y * y + xc; //z_{n+1}=(z_n)^2+c;
				y = 2 * x * y + yc;
				x = tempx;
				REAL r2 = x * x + y * y; //r2=|z_k|^2
				if ((r2 > 4) || k == niter - 1) //divergence condition
				{
					d_a[i + j * nElemX] = k;
					break;
				}
			}
}
////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	////////////////////////  Declarations ////////////////////////
	// pointers for host memory
	uint *h_a_cpu; //pointer for CPU results
	uint *h_a_gpu; //pointers for GPU results on host
	// pointer for device memory
	uint *d_a;//pointers for GPU results on device
	// define grid and block sizes
	int numBlocksX = 320;
	int numBlocksY = 320;
	int numThreadsPerBlockX = 32;
	int numThreadsPerBlockY = 8;
	// number of elements and memory size
	int nElemX = numBlocksX * numThreadsPerBlockX;
	int nElemY = numBlocksY * numThreadsPerBlockY;
	size_t memSize = nElemX * nElemY * sizeof(uint);
	// timers
	uint timer_gpu = 0;
	uint timer_cpu = 0;
	uint timer_mt = 0;
	//parameters for the mandelbrot set
	int niter = 100; //maximum number of iterations
	REAL xmin = -2; //limits for c=x+i*y
	REAL xmax = 1;
	REAL ymin = -1.5;
	REAL ymax = 1.5;
	////////////////////////  Allocate Memory   ////////////////////////
	//Allocate host memory
	h_a_cpu = (uint *) malloc(memSize);
	h_a_gpu = (uint *) malloc(memSize);
	//Allocate devie memory
	//TODO
	CUDA_SAFE_CALL(cudaMalloc ((void**)&d_a, memSize));
	// Check for any CUDA errors
	checkCUDAError("cudaMalloc");
	// create timers
	CUT_SAFE_CALL(cutCreateTimer(&timer_gpu));
	CUT_SAFE_CALL(cutCreateTimer(&timer_cpu));
	CUT_SAFE_CALL(cutCreateTimer(&timer_mt));
		CUT_SAFE_CALL(cutStartTimer(timer_cpu));
	////////////////////////  Calculation (host)  ////////////////////////
	for (int i = 0; i < nElemX; i = i + 1) //loop in the x-direction
	{
		REAL xc = xmin + (xmax - xmin) / (nElemX - 1) * i; //xc=real(c)
		for (int j = 0; j < nElemY; j = j + 1) //loop in the y-direction
		{
			REAL yc = ymin + (ymax - ymin) / (nElemY - 1) * j; //yc=imag(c)
			REAL x = 0.0; //x=real(z_k)
			REAL y = 0.0; //y=imag(z_k)
			for (int k = 0; k < niter; k = k + 1) //iteration loop
			{
				REAL tempx = x * x - y * y + xc; //z_{n+1}=(z_n)^2+c;
				y = 2 * x * y + yc;
				x = tempx;
				REAL r2 = x * x + y * y; //r2=|z_k|^2
				if ((r2 > 4) || k == niter - 1) //divergence condition
				{
					h_a_cpu[i + j * nElemX] = k;
					break;
				}
			}
		}
	}
	CUT_SAFE_CALL(cutStopTimer(timer_cpu));
	///////////////////////  Calculation (Device)            /////////////////////////
	CUT_SAFE_CALL(cutStartTimer(timer_gpu));
	dim3 dimGrid(numBlocksX, numBlocksY);
	dim3 dimBlock(numThreadsPerBlockX, numThreadsPerBlockY);
	mandelbrotKernel<<< dimGrid, dimBlock >>>(xmin, xmax, nElemX, ymin, ymax, nElemY, niter, d_a);
	// block until the device has completed
	cudaThreadSynchronize();
	// check if kernel execution generated an error
	checkCUDAError("kernel execution");
	CUT_SAFE_CALL(cutStopTimer(timer_gpu));
	////////////////////////  Copy Data From Device To Host   ////////////////////////
	//TODO
	// Check for any CUDA errors
	CUT_SAFE_CALL(cutStartTimer(timer_mt));
	cudaMemcpy (h_a_gpu, d_a, memSize, cudaMemcpyDeviceToHost);
	CUT_SAFE_CALL(cutStopTimer(timer_mt));
	checkCUDAError("cudaMemcpyDeviceToHost");
	////////////////////////  Compare the results             ////////////////////////
	for (int i = 0; i < nElemX; i++)
	{
		for (int j = 0; j < nElemY; j++)
		{
			//if (h_a_gpu[i + j * nElemX] != h_a_cpu[i + j * nElemX])
			if (abs (((long) h_a_gpu[i + j * nElemX]) - ((long)h_a_cpu[i + j * nElemX])) > 1) // allow difference up to 1 due to rounding
			{
				printf(
						"Error : i= %d j= %d h_a_gpu[i*nElemY+j]= %d h_a_cpu[i*nElemY+j ]= %d \n",
						i, j, h_a_gpu[i + j * nElemX], h_a_cpu[i + j * nElemX]);
						printf("\nPress ENTER to exit...\n");
						getchar();
				return 1;
			}
		}
	}
	////////////////////////  Elapsed time       /////////////////////////////////////////////
	REAL t_cpu =cutGetAverageTimerValue(timer_cpu);
	REAL t_gpu = cutGetAverageTimerValue(timer_gpu);
	REAL t_mt = cutGetAverageTimerValue(timer_mt);
	printf(
			"CUDA Mandelbrot niter= %d : GPU: % 4.1f ms MT: % 4.1f ms GPU+MT: % 4.1f ms CPU: % 4.1f ms speedup % 4.1f ( %4.1f )\n",
			niter, t_gpu, t_mt, t_gpu + t_mt, t_cpu, t_cpu / (t_gpu + t_mt),
			t_cpu / t_gpu);
	CUT_SAFE_CALL(cutResetTimer(timer_gpu));
	write_bmp("output.bmp", nElemX, nElemY, h_a_cpu);
        write_bmp("output_dev.bmp", nElemX, nElemY, h_a_gpu);

	////////////////////////  Free Memory      ////////////////////////
	// free device memory
	cudaFree(d_a);
	// free host memory
	free(h_a_cpu);
	free(h_a_gpu);
	//timers
	cutDeleteTimer(timer_gpu);
	////////////////////////  End    ////////////////////////
	printf("\nPress ENTER to exit...\n");
	getchar();
	return 0;
}

