#include "cutil.h"
#include <stdio.h>
#include <math_constants.h>
const int blockWidth = 512;
const int gridWidth = 100;
//######################################################################################
// Compare Matrices
//######################################################################################
void CompareArrays(const float * d_M, const float * h_M_cpu, const int nElem )
{
	float * h_M_gpu;
	const int size = nElem * sizeof(float);
	h_M_gpu = (float *) malloc(size);
	// Copy back results
	cudaMemcpy(h_M_gpu, d_M, size, cudaMemcpyDeviceToHost);
	for (int i = 0; i < nElem; i++)
	{
		float error = h_M_gpu[i] - h_M_cpu[i];
		if (abs(error) / (abs(h_M_cpu[i]) + 1) > 0.001)
		{
			fprintf(stderr, "Error: i= %d d_M[index]= %f h_M_cpu[index]= %f diff %f \n",
						i, h_M_gpu[i], h_M_cpu[i], error);
			free(h_M_gpu);
			return;
		}
	}
	free(h_M_gpu);
	return;
}
//######################################################################################
//  series calculation
//######################################################################################
__host__ __device__ float series_exp(const float x, const int niter)
{
	float res = 0.0;
	float xpower = 1.0;
	float kfact = 1;
	res += xpower / kfact;
	for (int k = 1; k < niter; k++)
	{
		xpower *= x;
		kfact *= k;
		if (isfinite (kfact))
			res += xpower / kfact;
	}
	return res;
}
__host__ float series_fourier_host(const float x, const int niter)
{
	float res = 0.0f;
	float m = -1.0f; // m = (-1) ^ (k-1);
	float ind = -1.0f; // ind = 2k - 1;
	for (int k = 1; k < niter; k++)
	{
		m *= -1.0f;
		ind += 2.0f;
		res += 4 / CUDART_PI_F * m * cos(ind*x) / ind;
	}
	return res;
}
__device__ float series_fourier_device(const float x, const int niter)
{
	float res = 0.0f;
	float m = -1.0f;
	float ind = -1.0f; // ind = 2k - 1;
	for (int k = 1; k < niter; k++)
	{
		m *= -1.0f;
		ind += 2.0f;
		//res +=4/CUDART_PI_F*m * cos(ind*x) / ind;
		//res += 4/CUDART_PI_F*m * __cosf(ind*x) / ind;
		res += 4 / CUDART_PI_F * m * __fdividef(__cosf(ind * x), ind);
	}
	return res;
}
void CalculationOnHost(float * h_M_cpu, const int nElem, const float xmin, const float xmax, const int niter)
{
	for (int i = 0; i < nElem; i++)
	{
		float x = xmin + ((xmax-xmin) / (nElem-1.0f)) * i;
		h_M_cpu[i] = series_fourier_host(x, niter);
	}
	return;
}
__global__ void CalculationOnDevice_Kernel(float * d_M, const float xmin, const float xmax, const int niter)
{
	int idx = blockIdx.x * blockDim.x + threadIdx.x;
	float x = xmin + ((xmax-xmin) / (blockDim.x*gridDim.x-1)) * idx;
	d_M[idx] = series_fourier_device(x, niter);
	return;
}
void CalculationOnDevice(float * d_M, const float xmin, const float xmax, const int niter)
{
	dim3 dimBlock(blockWidth, 1);
	dim3 dimGrid(gridWidth, 1);
	// Launch the device computation threads
	CalculationOnDevice_Kernel<<<dimGrid, dimBlock>>>(d_M, xmin, xmax, niter);
	CUDA_SAFE_CALL(cudaThreadSynchronize());
	return;
}
int main(int argc, char** argv)
{
	const int nElem = blockWidth * gridWidth;
	const float xmin = -10.0;
	const float xmax = 10.0;
	const int niter = 100;
	const int niter_timer = 1;
	unsigned int timer_cpu = 0;
	unsigned int timer_gpu = 0;
	CUT_SAFE_CALL(cutCreateTimer(&timer_cpu));
	CUT_SAFE_CALL(cutCreateTimer(&timer_gpu));
	// Allocate the array on the host
	float * h_M_cpu = (float *) malloc (nElem * sizeof (float));
	// Allocate the array on the device
	float * d_M;
	cudaMalloc((void**) &d_M, nElem * sizeof(float));
	//Calculation on the host
	CUT_SAFE_CALL(cutStartTimer(timer_cpu));
	for (int i = 0; i < niter_timer; i++)
	{
		CalculationOnHost(h_M_cpu, nElem, xmin, xmax, niter);
	}
	CUT_SAFE_CALL(cutStopTimer(timer_cpu));
	// Calculation on the device
	CUT_SAFE_CALL(cutStartTimer(timer_gpu));
	for (int i = 0; i < niter_timer; i++)
	{
		CalculationOnDevice(d_M, xmin, xmax, niter);
	}
	CUT_SAFE_CALL(cutStopTimer(timer_gpu));
	// Output results
	if (0)
	{
		for (int i = 0; i < nElem; i++)
		{
			float x = xmin + ((xmax-xmin) / (nElem-1.0f)) * i;
			printf ("%.10e %.10e\n", x, h_M_cpu[i]);
		}
	}
	//Compare results
	CompareArrays(d_M, h_M_cpu, nElem);
	// Free matrices on device
	CUDA_SAFE_CALL(cudaFree(d_M));
	free(h_M_cpu);
	// Output timer values
	float t_cpu = cutGetAverageTimerValue(timer_cpu);
	float t_gpu = cutGetAverageTimerValue(timer_gpu);
	fprintf(stderr, "Timer: CPU %f ms   GPU %f ms   speedup %f \n", t_cpu,
			t_gpu,  t_cpu / t_gpu);
	fprintf(stderr, "\nPress ENTER to exit...\n");
	getchar();
	return 0;
}

