#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#include <curand.h>
#include <cutil_inline.h>

#define N 20480
#define THREADS_PER_BLOCK 256

//CPU Approximation 
float piEstimateCPU(float *x, float *y)
{
    int s=0;
    for( int i = 0; i < N; i++)
    {
    	float xx = x[i];
		float yy = y[i];
		float norm = xx*xx + yy*yy;
		if (norm < 1.0f)
			s++;
    }
    return 4.0*s/N;
}

__global__ void piEstimateGPU(float *x_dev,float *y_dev, float *p_dev){
	
	int idx=blockIdx.x*blockDim.x+threadIdx.x;
	__shared__ int temp[THREADS_PER_BLOCK];	
	
	if(idx> N) return;
	
	float xx=x_dev[idx];
	float yy=y_dev[idx];
	float norm =xx*xx+yy*yy;
	if (norm <1.0f)  temp[threadIdx.x]=1;
	else temp[threadIdx.x]=0;
	__syncthreads();
	
	if(threadIdx.x==0){	
		int count=0;
		for(int i=0;i<THREADS_PER_BLOCK;i++) count+=temp[i];
		float pi=4.0*count/N;
		atomicAdd(p_dev,pi);
	}
		
	
}

int main(int argc, char *argv[])
{
	float *x, *y,*x_dev,*y_dev, *p_dev;
	float p=0.0;
	float p_cpu=0.0;
	curandGenerator_t gen;
	
        
	//Define grid dimension and block dimension
	dim3 mblockDim(THREADS_PER_BLOCK);
	dim3 mgridDim(N/THREADS_PER_BLOCK+1);
	
        // Memory allocation on CPU
	x     = (float *)malloc(N*sizeof(float));
    	y     = (float *)malloc(N*sizeof(float));
	
	
    	// Memory allocation on GPU
    	cudaMalloc((void **)&x_dev, N*sizeof(float));
    	cudaMalloc((void **)&y_dev, N*sizeof(float));
	cudaMalloc((void **)&p_dev, sizeof(float));
   
	// Create pseudo-random number generator
	curandCreateGenerator(&gen,CURAND_RNG_PSEUDO_DEFAULT);
	// Set seed
	curandSetPseudoRandomGeneratorSeed(gen, 1234ULL);
	// Generate x and y on device 
	curandGenerateUniform(gen, x_dev, N);
	curandGenerateUniform(gen, y_dev, N);

	
	//Transfer memory from  host to device
	cudaMemcpy(x,x_dev,N*sizeof(float),cudaMemcpyDeviceToHost);
	cudaMemcpy(y,y_dev,N*sizeof(float),cudaMemcpyDeviceToHost);
	cudaMemcpy(p_dev,&p,sizeof(float),cudaMemcpyHostToDevice);
	
	
	//Call kernel
	piEstimateGPU<<<mgridDim,mblockDim>>>(x_dev,y_dev,p_dev);
	cudaMemcpy(&p,p_dev,sizeof(float),cudaMemcpyDeviceToHost);
	printf("GPU pi=%f\n", p);


	p_cpu=piEstimateCPU(x, y);
	printf("CPU pi=%f\n", p_cpu);

	// Cleanup 
	curandDestroyGenerator(gen);
	cudaFree(x_dev);
	cudaFree(y_dev);
	cudaFree(p_dev);
	free(x);
	free(y);	
	
}

